/* --------------------------------------------------------------------------------
 * Example sketch/program showing how to read data from a PICC to serial.
 * This is modified from a MFRC522 library example; 
 * For project 3 in course EITF55
 * Ben Smeets 2018-01-06
 * @license Released into the public domain.
*/
#include <SPI.h>
#include <MFRC522.h>

constexpr uint8_t RST_PIN = 9;
constexpr uint8_t SS_PIN = 10;
MFRC522 mfrc522(SS_PIN, RST_PIN);  // Create MFRC522 instance

MFRC522::MIFARE_Key key;

void setup() {
	Serial.begin(9600);		// Initialize serial communications with the PC
	while (!Serial);		// Do nothing if no serial port is opened 
	SPI.begin();			// Init SPI bus
	mfrc522.PCD_Init();		// Init MFRC522
	mfrc522.PCD_DumpVersionToSerial();	// Show details of PCD - MFRC522 Card Reader
	Serial.println(F("Scan PICC to see UID, SAK, type, and data blocks..."));

    // Prepare the key (used both as key A and as key B)
    // using FFFFFFFFFFFFh which is the default at chip delivery from the factory
    for (byte i = 0; i < 6; i++) {
        key.keyByte[i] = 0xFF;
    }
}

/*  Main loop */
void loop() {
    // Look for new cards
    if ( ! mfrc522.PICC_IsNewCardPresent())
        return;

    // Select one of the cards
    if ( ! mfrc522.PICC_ReadCardSerial())
        return;

    // Show some details of the PICC (that is: the tag/card)
    Serial.print(F("Card UID:")); dump_byte_array(mfrc522.uid.uidByte, mfrc522.uid.size);
    Serial.println();

    // here comes your fancy code ===============================================================
    // For example, we read from: block 0 in sector #1, i.e. block 4
    byte sector         = 1;
    byte blockAddr      = 4;
    byte dataBlock[]    = {
        0x01, 0x02, 0x03, 0x04, //  1,  2,   3,  4,
        0x05, 0x06, 0x07, 0x08, //  5,  6,   7,  8,
        0x09, 0x0a, 0xff, 0x0b, //  9, 10, 255, 11,
        0x0c, 0x0d, 0x0e, 0x0f  // 12, 13, 14, 15
    };
    byte trailerBlock   = 7;
    MFRC522::StatusCode status;
    
    // buffer to receive the 16 data bytes of a block + 2 check bytes
    byte buffer[18]; 
    byte size = sizeof(buffer);

    // Authenticate using key A
    Serial.println(F("Authenticating using key A..."));
    status = (MFRC522::StatusCode) mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_A, trailerBlock, &key, &(mfrc522.uid));
    if (status != MFRC522::STATUS_OK) {
        myErrorPrint("PCD_Authenticate() failed: ", status);
        return;
    }

    // Read data from the block
    status = (MFRC522::StatusCode) mfrc522.MIFARE_Read(blockAddr, buffer, &size);
    if (status != MFRC522::STATUS_OK) {
        myErrorPrint("MIFARE_Read() failed: ",status);
        return;
    }
    SerialBlockPrint("Data in block ", blockAddr, buffer); 

    // Authenticate using key B
    Serial.println(F("Authenticating again using key B..."));
    status = (MFRC522::StatusCode) mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_B, trailerBlock, &key, &(mfrc522.uid));
    if (status != MFRC522::STATUS_OK) {
        myErrorPrint("PCD_Authenticate() failed: ", status);
        return;
    }

    // Write data to the block
    SerialBlockPrint("Writing data into block ", blockAddr, dataBlock);
    
    status = (MFRC522::StatusCode) mfrc522.MIFARE_Write(blockAddr, dataBlock, 16);
    if (status != MFRC522::STATUS_OK) {
        myErrorPrint("MIFARE_Write() failed: ", status);
        return;
    }

    // Dump the sector data
    Serial.println(F("Current data in sector:"));
    mfrc522.PICC_DumpMifareClassicSectorToSerial(&(mfrc522.uid), &key, sector);
    Serial.println();

    // END OF Example =============================================================
    // Halt PICC
    mfrc522.PICC_HaltA();
    // Stop encryption on PCD, always do this if authentication was used !!!
    mfrc522.PCD_StopCrypto1();
    } /* loop ends here */

/**
 * Helper routines 
 */    
void myErrorPrint(char *msg, MFRC522::StatusCode status) {
        Serial.print(msg);
        Serial.println(mfrc522.GetStatusCodeName(status));
 }

void  SerialBlockPrint(char *msg, byte addr, byte *buf) {
    Serial.print(msg); Serial.print(addr);
    Serial.println(F(" ..."));
    dump_byte_array(buf, 16); Serial.println();  
}

void dump_byte_array(byte *buffer, byte bufferSize) {
    for (byte i = 0; i < bufferSize; i++) {
        Serial.print(buffer[i] < 0x10 ? " 0" : " ");
        Serial.print(buffer[i], HEX);
    }
}
/**
 * Use this routine to continue program after an error occurred.
 * You may have to reauthenticate again if the error occorred in
 * an authenticated state.
 */
bool ErrorRecover(){
      MFRC522::StatusCode status;
      byte bufferATQA[10];
      byte bufferSize[10];
      mfrc522.PICC_HaltA();
      // Stop encryption on PCD, always do this if authentication was used !!!
      mfrc522.PCD_StopCrypto1();
      delay(500); 
      if ( status = mfrc522.PICC_WakeupA(bufferATQA, bufferSize)){ //wake up PICC which was previously Halted
        Serial.println(mfrc522.GetStatusCodeName(status));
        return false;
      }
      delay(500);   
      // Select the card again
      if ( ! mfrc522.PICC_ReadCardSerial()){
        return false;
      }
      Serial.println("Recovered");
      return true;
}